/*! \file meanvalue.cpp
 *  \brief This file contains the implementation of mean value coordinates for closed triangular meshes
 *  \author Albert Busqué
 *
 * The code show here is based on the following paper:
 * http://www.cs.wustl.edu/~taoju/research/meanvalue.pdf
*/

#include "meanvalue.h"
#include "const.h"
#include "utils.h"

/*! 
 * \brief Compute mean value coordinates of a point.
 *
 * This function is called internally by ::visualize
 *
 * \param[in] ca	A struct Cage identifying cage points and faces
 * \param[in] x		3-dim point on which we calculate the coordinates
 *
 * \param[out] w	Mean value coordinates  of x point
 *
 * This function computes mean value coordinates of a point in array x
 * and stores it in array w (the length of w is equal to the number of 
 * points in the box).
*/
void compute_coordinates (Cage *ca, double *x, float *w)
{	
	int sign;
	unsigned int nt = ca->nfaces, npc = ca->npts, i, k, next, prev, p, inext, iprev, jump = FALSE;
	double dist[npc], u[npc][3], l[3], theta[3], h, c[3], s[3], res, sum = 0.0, wi, ll;

	// Per cada vertex de la caixa calculem el modul de la distancia entre cada punt de la caixa i el punt x
	for (i=0; i<npc; i++)
	{
		dist[i] = sqrt(pow(ca->points[i*3+0]-x[0],2.0) + pow(ca->points[i*3+1]-x[1],2.0) + pow(ca->points[i*3+2]-x[2],2.0));	// Calcul modul
		
		if (dist[i] < TOL)	// Si el punt es molt a prop de la caixa assignem 1 al pes corresponent i la resta 0's 
		{		
			w[i] = 1.0;
			return;
		}
		
		// Normalitzem el vector distancia pel seu modul
		u[i][0] = (ca->points[i*3+0]-x[0]) / dist[i];
		u[i][1] = (ca->points[i*3+1]-x[1]) / dist[i];
		u[i][2] = (ca->points[i*3+2]-x[2]) / dist[i];
	}

	for (k=0; k<nt; k++)	// Per cada triangle de la caixa
	{		
		h = 0.0;
		for (i=0; i<3; i++)	//Per cada vertex de cada triangle
		{		 	
			// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
			next = (i+1+3) % 3;
			prev = (i-1+3) % 3;

			// Calculem el modul de la distancia entre el punt seguent menys l'anterior
			ll = 0.0;
			for (p=0; p<3; p++)
			{
				inext = ca->faces[k*3+next];
				iprev = ca->faces[k*3+prev];
				ll += pow(u[inext][p] - u[iprev][p], 2.0);
			}

			l[i] = sqrt(ll);

			// Calculem l'angle theta
			theta[i] = 2.0 * asin((l[i]/2.0));
			
			//Calculem h
			h += theta[i] / 2.0; 
		}

		// Preguntem si el punt x es sobre el triangle
		if ((M_PI - h) < TOL)
		{			
			// Netejem els pesos
			for (i=0; i<npc; i++)
				w[i] = 0.0;

			// Netejem el sumatori
			sum = 0.0;

			// Anem a calcular els pesos wi
			for (i=0; i<3; i++)
			{
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				// Computem els pesos
				wi = sin(theta[i]) * l[prev] * l[next];
				w[ca->faces[k*3+i]] += wi; 			
				sum += wi;
			}
		
			// Sortim de l'evaluacio dels triangles, anem a normalitzar els pesos i retornar de la funcio
			break;
		}
		else	// Ja tenim h calculada. Anem a calcular c, s i w. Tornem a recorrer els vertex del triangle
		{	
			for (i=0; i<3; i++)
			{				
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				c[i] = ((2.0*sin(h)*sin(h-theta[i])) / (sin(theta[next])*sin(theta[prev]))) - 1.0;

				// Calculem el signe del determinant
				res = determinant3x3(u[ca->faces[k*3+0]], u[ca->faces[k*3+1]], u[ca->faces[k*3+2]]);
				if (res >= 0.0)
					sign = 1;
				else
					sign = -1;

				s[i] = sign * sqrt(1-(c[i]*c[i]));
				
				// Preguntem si x es fora del mateix pla. En aquest cas ignorem el triangle 
				if ( fabs(s[i]) <= TOL )
				{
					jump = TRUE;
					break;
				}
			}

			// Preguntem si s'ha d'ignorar el triangle
			if (jump){
				jump = FALSE;
				continue;
			}

			// Anem a calcular els pesos wi
			for (i=0; i<3; i++)
			{
				// Calculem els idx dels vertex posterior i anterior al actual a l'array de triangles
				next = (i+1+3) % 3;
				prev = (i-1+3) % 3;

				// Computem els pesos
				wi = (theta[i] - (c[next]*theta[prev]) - (c[prev]*theta[next])) / (dist[ca->faces[k*3+i]] * sin(theta[next]) * s[prev]);
				w[ca->faces[k*3+i]] += wi; 			
				sum += wi;
			}
		}
	}

	// Normalitzem els pesos
	for(i=0; i<npc; i++)
		w[i] /= sum;
}

void free_cage (Cage *c)
{
	free(c->faces);
}
