/*! \file visualization.cpp
 *  \brief Visualization of data
 *  \author Albert Busqué
 * 
 * This file contains functions and classes to display the cage and figure.
 * It uses VTK libraries.
*/

#include "visualization.h"
#include "MouseInteractorStylePP.h"
#include "parallel.h"
#include "utils.h"
#include <sys/time.h>

void create_polydata_cube_cage (vtkSmartPointer<vtkActor> fig, vtkSmartPointer<vtkPolyData> polydata)
{
	double bounds[6];	
	int k, i, j;
	unsigned int faces[12][3] = {{0,1,2}, {2,1,6}, {2,3,0},{0,3,4}, {4,3,7}, {0,4,5}, {0,5,1}, {1,5,6}, {6,5,7}, {6,7,3}, {6,3,2}, {4,7,5}};

	// Getting bounds of figure as (Xmin,Xmax,Ymin,Ymax,Zmin,Zmax)
	fig->GetBounds(bounds);

	// Expand margins of cage	
	for(k=0; k<6; k=k+2)
  	{
    		bounds[k] = bounds[k] - 0.001;
  	}
	for(k=1; k<6; k=k+2)
  	{
    		bounds[k] = bounds[k] + 0.001;
  	}

	// Create and fill vtkPoints array with points of cage
	vtkSmartPointer<vtkPoints> points = vtkSmartPointer<vtkPoints>::New();
	points->InsertNextPoint (bounds[0], bounds[2], bounds[4]);
 	points->InsertNextPoint (bounds[1], bounds[2], bounds[4]);
  	points->InsertNextPoint (bounds[1], bounds[3], bounds[4]);
  	points->InsertNextPoint (bounds[0], bounds[3], bounds[4]);
  	points->InsertNextPoint (bounds[0], bounds[2], bounds[5]);
  	points->InsertNextPoint (bounds[1], bounds[2], bounds[5]);
  	points->InsertNextPoint (bounds[1], bounds[3], bounds[5]);
  	points->InsertNextPoint (bounds[0], bounds[3], bounds[5]);

	// Create array of faces (geometry) and topology
	vtkSmartPointer<vtkCellArray> topology = vtkSmartPointer<vtkCellArray>::New();
	for (i=0; i<12; i++)
	{
		/* Create each face. Each triangle is formed by three points.
		First parameter of SetId is the index of the triangle vertex which is ALWAYS 0-2.
  		Second parameter of SetId is the index into the point (geometry) array, so this can range from 0-(NumPoints-1) */
		vtkSmartPointer<vtkTriangle> triangle = vtkSmartPointer<vtkTriangle>::New();
		triangle->GetPointIds()->SetNumberOfIds(3);
		for (j=0; j<3; j++)
		{
			triangle->GetPointIds()->SetId ( j, faces[i][j] );
		}
		topology->InsertNextCell (triangle);
	}

  	// Add the geometry and topology to the polydata
  	polydata->SetPoints ( points );
  	polydata->SetPolys ( topology );
}

void create_cage_from_polydata (vtkSmartPointer<vtkPolyData> polydata, Cage *c)
{
	unsigned int i, j;

	// Set number of box points
	c->npts = polydata->GetNumberOfPoints();

	// Set points
	c->points = reinterpret_cast<float*>(polydata->GetPoints()->GetVoidPointer(0));	
	
	// Set number of faces
	c->nfaces = polydata->GetNumberOfCells();

	// Allocate memory for these
	c->faces = (unsigned int*) malloc(sizeof(unsigned int ) * c->nfaces * 3);
	if (!c->faces)
	{
		printf("Error: No memory available\n");
		exit(EXIT_FAILURE);
	}

	for (i=0; i<c->nfaces; i++)
		for (j=0; j<3; j++)
		{
			c->faces[i*3+j] =  polydata->GetCell(i)->GetPointId(j);
		}
}

void read_model_from_file (char* filename, vtkSmartPointer<vtkPolyDataMapper> mapper) 
{
	char *ext;

	// Check if figure file exists
	if ( !file_exists(filename) )
	{
		printf("Error: file %s not exists\n", filename);
		exit(EXIT_FAILURE);
	}	

	// Read figure file extension and store it to 'ext'
	ext = get_extension (filename);

	// Read PLY file containing figure
	if (strncmp(ext+1, "ply", 3) == 0)
	{
		vtkSmartPointer<vtkPLYReader> ply_reader = vtkSmartPointer<vtkPLYReader>::New();
		ply_reader->SetFileName (filename);

		// Setup actor and mapper
		#if VTK_MAJOR_VERSION <= 5
  			mapper->SetInput(ply_reader->GetOutput());
		#else
  			mapper->SetInputData(ply_reader->GetOutput());
		#endif
	}
	// Or read OBJ file containing figure
	else if (strncmp(ext+1, "obj", 3) == 0)
	{
		vtkSmartPointer<vtkOBJReader> obj_reader = vtkSmartPointer<vtkOBJReader>::New();
		obj_reader->SetFileName (filename);

		// Setup mapper
		#if VTK_MAJOR_VERSION <= 5
  			mapper->SetInput(obj_reader->GetOutput());
		#else
  			mapper->SetInputData(obj_reader->GetOutput());
		#endif
	}
	// If no known extension
	else
	{
		printf("Error: %s is not valid extension. Valid extensions: OBJ, PLY.\n", ext);
		exit(EXIT_FAILURE);
	}

	mapper->Update();
}

/*! 
 * \brief Display data 
 *
 * This function is called internally by ::main
 *
 * This function do all the work to start visualization and interaction.
 * Loads figure and cage objects from file.
 * Also coordinates are computed once loaded the cage and figure.
 *
 * See also ::compute_coordinates
*/
void visualize (struct global_args args)
{
	unsigned int fig_npts;
	float *mvcoords; 
	Cage cage;
	struct cl_data clData;

	//---------------------------------------------------------------------------------------------------- Loading figure

	// Create a polydata mapper for figure
	vtkSmartPointer<vtkPolyDataMapper> mapper_figure = vtkSmartPointer<vtkPolyDataMapper>::New();

	read_model_from_file(args.figure_filename, mapper_figure);

	// Setup actor
 	vtkSmartPointer<vtkActor> actor_figure = vtkSmartPointer<vtkActor>::New();
  	actor_figure->SetMapper(mapper_figure);
	//mapper_figure->ImmediateModeRenderingOn();
	
	//---------------------------------------------------------------------------------------------------- Loading cage

	// Create a polydata mapper for cage
	vtkSmartPointer<vtkPolyDataMapper> mapper_cage = vtkSmartPointer<vtkPolyDataMapper>::New();
	
	// If specified any cage filename then create default cube cage from bounds of figure
	if (!args.cage_filename)
	{	
		vtkSmartPointer<vtkPolyData> cage_polydata = vtkSmartPointer<vtkPolyData>::New();

		create_polydata_cube_cage(actor_figure, cage_polydata);
			
		#if VTK_MAJOR_VERSION <= 5
			mapper_cage->SetInput(cage_polydata);
		#else
			mapper_cage->SetInputData(cage_polydata);
		#endif
	}
	// Otherwise we create cage from specified file			
	else
	{
		read_model_from_file(args.cage_filename, mapper_cage);
	}

	// Fill cage structure with polydata
	create_cage_from_polydata(mapper_cage->GetInput(), &cage);

	// Setup actor
	vtkSmartPointer<vtkActor> actor_cage = vtkSmartPointer<vtkActor>::New();
	actor_cage->SetMapper(mapper_cage);
		
	//---------------------------------------------------------------------------------------------------- Computing mv coordinates
	
	// Compute mean value coordinates of figure
	fig_npts = mapper_figure->GetInput()->GetNumberOfPoints();
	mvcoords = (float *) calloc(fig_npts * cage.npts, sizeof(float));
	if (!mvcoords)
	{
		printf("Error: Not memory enough.\n");
		exit(EXIT_FAILURE);
	}

	struct timeval start, end;
	long mtime, seconds, useconds;	

	#ifndef OPENCL
		double p[3];
		unsigned int i;
		
		gettimeofday(&start, NULL);

		for (i=0; i<fig_npts; i++)
		{ 
			mapper_figure->GetInput()->GetPoint(i, p);
			compute_coordinates(&cage, p, mvcoords+(i*cage.npts));
		}

		gettimeofday(&end, NULL);

		seconds  = end.tv_sec  - start.tv_sec;
    		useconds = end.tv_usec - start.tv_usec;

		mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Elapsed time MVC: %ld milliseconds\n", mtime);
	#else
		float* figure_points = reinterpret_cast<float*>(mapper_figure->GetInput()->GetPoints()->GetVoidPointer(0));

		// Use this to check the output of each API call
		cl_int status; 
		
		init_opencl_resources (&clData);

		gettimeofday(&start, NULL);

		compute_parallel_coordinates (&clData, &cage, fig_npts, figure_points, mvcoords);

		gettimeofday(&end, NULL);

		seconds  = end.tv_sec  - start.tv_sec;
    		useconds = end.tv_usec - start.tv_usec;

		mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Elapsed time MVC: %ld milliseconds\n", mtime);
	
		
		//-----------------------------------------------------
		// STEP 8: Create the kernel
		//----------------------------------------------------- 

		// Use clCreateKernel() to create a kernel from the 
		// vector addition function (named "vecadd")
		clData.kernel = clCreateKernel(clData.program, "recover_point", &status);
		chk(status, "clCreateKernel");	

		//-----------------------------------------------------
		// STEP 9: Set the kernel arguments
		//----------------------------------------------------- 

		// Associate the input and output buffers with the 
		// kernel 
		// using clSetKernelArg()
		status  = clSetKernelArg(clData.kernel, 0, sizeof(cl_mem), &(clData.bufferMV));
		status |= clSetKernelArg(clData.kernel, 1, sizeof(cl_mem), &(clData.bufferCage));
		status |= clSetKernelArg(clData.kernel, 2, sizeof(int), &cage.npts);
		status |= clSetKernelArg(clData.kernel, 3, sizeof(cl_mem), &(clData.bufferFigure));
		chk(status, "clSetKernelArg");
	
	#endif

	//---------------------------------------------------------------------------------------------------- Starting visualization

  	// Create a renderer, render window, and an interactor
  	vtkSmartPointer<vtkRenderer> renderer = vtkSmartPointer<vtkRenderer>::New();
  	vtkSmartPointer<vtkRenderWindow> renderWindow = vtkSmartPointer<vtkRenderWindow>::New();
  	renderWindow->AddRenderer(renderer);
  	vtkSmartPointer<vtkRenderWindowInteractor> renderWindowInteractor = vtkSmartPointer<vtkRenderWindowInteractor>::New();
  	renderWindowInteractor->SetRenderWindow(renderWindow);

	// Create Point Picker
	vtkSmartPointer<vtkPointPicker> pointPicker = vtkSmartPointer<vtkPointPicker>::New();
 	renderWindowInteractor->SetPicker(pointPicker);

	// Set style
	vtkSmartPointer<MouseInteractorStylePP> style = vtkSmartPointer<MouseInteractorStylePP>::New();
	style->SetPolyDataCage(mapper_cage->GetInput());
	style->SetPolyDataFigure(mapper_figure->GetInput());
	style->SetMeanValueCoords(mvcoords);
	style->SetClData(&clData);
	style->SetCageRef(actor_cage);
	style->SetCurrentRenderer(renderer);
	renderWindowInteractor->SetInteractorStyle(style);
	
  	// Add the actors to the scene
  	renderer->AddActor(actor_cage);
	renderer->AddActor(actor_figure);
  	renderer->SetBackground(0.1804,0.5451,0.3412);		// Background color green

  	// Render and interact
	renderWindow->Render();
	renderWindowInteractor->Start();

	// Free data
	#ifdef OPENCL

		//-----------------------------------------------------
		// STEP 13: Release OpenCL resources
		//----------------------------------------------------- 

		// Free OpenCL resources
		clReleaseKernel(clData.kernel);
		clReleaseProgram(clData.program);
		clReleaseCommandQueue(clData.cmdQueue);
		clReleaseMemObject(clData.bufferMV);
		clReleaseMemObject(clData.bufferCage);
		clReleaseMemObject(clData.bufferFigure);
		clReleaseContext(clData.context);
		
		// Free host resources
		free(clData.platforms);
		free(clData.devices);

	#endif
		free_cage (&cage);
		free(mvcoords);	
}
